import os.path
import sys
import codecs
import json
import functools
import six
from six.moves import urllib

try:
    from urlparse import urljoin  # Python2
except ImportError:
    from urllib.parse import urljoin  # Python3
pathname2url = urllib.request.pathname2url
from jsonschema import RefResolver

from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap
from ruamel.yaml.representer import RepresenterError, SafeRepresenter
from ruamel.yaml.constructor import ConstructorError

from .util import (
    to_bytes,
    to_text,
    sensitive_str,
    sensitive_bytes,
    sensitive_dict,
    sensitive_list,
    wrapSensitiveValue,
    UnfurlError,
    UnfurlValidationError,
    findSchemaErrors,
)
from .merge import (
    expandDoc,
    makeMapWithBase,
    findAnchor,
    _cacheAnchors,
    restoreIncludes,
)

from toscaparser.common.exception import URLException, ExceptionCollector
from toscaparser.utils.gettextutils import _

from ansible.parsing.vault import VaultLib, VaultSecret
from ansible.utils.unsafe_proxy import AnsibleUnsafeText, AnsibleUnsafeBytes

import logging

logger = logging.getLogger("unfurl")


def represent_undefined(self, data):
    raise RepresenterError(
        "cannot represent an object: <%s> of type %s" % (data, type(data))
    )


def _represent_sensitive(dumper, data, tag):
    if dumper.vault.secrets:
        b_ciphertext = dumper.vault.encrypt(data)
        return dumper.represent_scalar(tag, b_ciphertext.decode(), style="|")
    else:
        return dumper.represent_scalar(
            u"tag:yaml.org,2002:str", sensitive_str.redacted_str
        )


def represent_sensitive_str(dumper, data):
    return _represent_sensitive(dumper, data, u"!vault")


def represent_sensitive_json(dumper, data):
    return _represent_sensitive(
        dumper, json.dumps(data, sort_keys=True), u"!vault-json"
    )


def represent_sensitive_bytes(dumper, data):
    return _represent_sensitive(dumper, data, u"!vault-binary")


def _construct_vault(constructor, node, tag):
    value = constructor.construct_scalar(node)
    if not constructor.vault.secrets:
        raise ConstructorError(
            context=None,
            context_mark=None,
            problem="found %s but no vault password provided" % tag,
            problem_mark=node.start_mark,
            note=None,
        )
    return value


def construct_vault(constructor, node):
    value = _construct_vault(constructor, node, "!vault")
    cleartext = to_text(constructor.vault.decrypt(value))
    return sensitive_str(cleartext)


def construct_vaultjson(constructor, node):
    value = _construct_vault(constructor, node, "!vault-json")
    cleartext = to_text(constructor.vault.decrypt(value))
    return wrapSensitiveValue(json.loads(cleartext))


def construct_vaultbinary(constructor, node):
    value = _construct_vault(constructor, node, "!vault-binary")
    cleartext = constructor.vault.decrypt(value)
    return sensitive_bytes(cleartext)


def makeYAML(vault=None):
    if not vault:
        vault = VaultLib(secrets=None)
    yaml = YAML()

    # monkey patch for better error message
    yaml.representer.represent_undefined = represent_undefined
    yaml.representer.add_representer(None, represent_undefined)

    yaml.constructor.vault = vault
    yaml.constructor.add_constructor(u"!vault", construct_vault)
    yaml.constructor.add_constructor(u"!vault-json", construct_vaultjson)
    yaml.constructor.add_constructor(u"!vault-binary", construct_vaultbinary)

    yaml.representer.vault = vault
    yaml.representer.add_representer(sensitive_str, represent_sensitive_str)
    yaml.representer.add_representer(sensitive_dict, represent_sensitive_json)
    yaml.representer.add_representer(sensitive_list, represent_sensitive_json)
    yaml.representer.add_representer(sensitive_bytes, represent_sensitive_bytes)

    if six.PY3:
        represent_unicode = SafeRepresenter.represent_str
        represent_binary = SafeRepresenter.represent_binary
    else:
        represent_unicode = SafeRepresenter.represent_unicode
        represent_binary = SafeRepresenter.represent_str

    yaml.representer.add_representer(AnsibleUnsafeText, represent_unicode)
    yaml.representer.add_representer(AnsibleUnsafeBytes, represent_binary)
    return yaml


yaml = makeYAML()


def makeVaultLib(passwordBytes, vaultId="default"):
    if passwordBytes:
        if isinstance(passwordBytes, six.string_types):
            passwordBytes = to_bytes(passwordBytes)
        vault_secrets = [(vaultId, VaultSecret(passwordBytes))]
        return VaultLib(secrets=vault_secrets)
    return None


def resolveIfInRepository(manifest, path, isFile, importLoader=None):
    # check if this path is a git url, if it is, return the repo and local path to the file
    repo, filePath, revision, bare = manifest.findRepoFromGitUrl(
        path, isFile, importLoader, True
    )
    if repo:  # it's a git repo
        if bare:
            if not filePath:
                raise UnfurlError(
                    "can't retrieve local repository for '%s' with revision '%s'"
                    % (path, revision)
                )
            contents = repo.show(filePath, revision)
            return filePath, six.StringIO(contents), True
        path = os.path.join(repo.workingDir, filePath)
        isFile = True
    else:
        # if it's a file path, check if it's in one of our repos
        if isFile:
            repo, filePath, revision, bare = manifest.findPathInRepos(
                path, importLoader, True
            )
            if repo:
                assert not bare
                path = os.path.join(repo.workingDir, filePath)
    return path, None, isFile


_refResolver = RefResolver("", None)


def load_yaml(yaml, path, isFile=True, importLoader=None, fragment=None):
    try:
        logger.debug(
            "attempting to load YAML %s: %s", "file" if isFile else "url", path
        )
        originalPath = path
        f = None
        manifest = importLoader and getattr(importLoader.tpl, "manifest", None)
        if manifest:
            path, f, isFile = resolveIfInRepository(
                manifest, path, isFile, importLoader
            )

        if not f:
            try:
                if isFile:
                    ignoreFileNotFound = importLoader and getattr(
                        importLoader.tpl, "ignoreFileNotFound", None
                    )
                    if ignoreFileNotFound and not os.path.isfile(path):
                        return None

                    if manifest and manifest.loader:
                        # show == True if file was decrypted
                        contents, show = manifest.loader._get_file_contents(path)
                        f = six.StringIO(codecs.decode(contents))
                    else:
                        f = codecs.open(path, encoding="utf-8", errors="strict")
                else:
                    f = urllib.request.urlopen(path)
            except urllib.error.URLError as e:
                if hasattr(e, "reason"):
                    msg = _(
                        'Failed to reach server "%(path)s". Reason is: ' "%(reason)s."
                    ) % {"path": path, "reason": e.reason}
                    ExceptionCollector.appendException(URLException(what=msg))
                    return
                elif hasattr(e, "code"):
                    msg = _(
                        'The server "%(path)s" couldn\'t fulfill the request. '
                        'Error code: "%(code)s".'
                    ) % {"path": path, "code": e.code}
                    ExceptionCollector.appendException(URLException(what=msg))
                    return
                else:
                    raise
        with f:
            doc = yaml.load(f.read())
            if isinstance(doc, CommentedMap):
                doc.path = path
        if fragment:
            return _refResolver.resolve_fragment(doc, fragment)
        else:
            return doc
    except:
        if path != originalPath:
            msg = 'Could not load "%s" (originally "%s")' % (path, originalPath)
        else:
            msg = 'Could not load "%s"' % path
        raise UnfurlError(msg, True)


import toscaparser.imports

toscaparser.imports.YAML_LOADER = functools.partial(load_yaml, yaml)


def loadYamlFromArtifact(context, artifact, yaml):
    # _load_import_template will invoke load_yaml above
    loader = toscaparser.imports.ImportsLoader(
        None,
        artifact.baseDir,
        tpl=context,
        yaml_loader=functools.partial(load_yaml, yaml),
    )
    return loader._load_import_template(None, artifact.asImportSpec())


class YamlConfig(object):
    def __init__(
        self,
        config=None,
        path=None,
        validate=True,
        schema=None,
        loadHook=None,
        vault=None,
    ):
        try:
            self._yaml = None
            self.vault = vault
            self.path = None
            self.schema = schema
            if path:
                self.path = os.path.abspath(path)
                if os.path.isfile(self.path):
                    with open(self.path, "r") as f:
                        config = f.read()
                # otherwise use default config
            else:
                self.path = None

            if isinstance(config, six.string_types):
                if self.path:
                    # set name on a StringIO so parsing error messages include the path
                    config = six.StringIO(config)
                    config.name = self.path
                self.config = self.yaml.load(config)
            elif isinstance(config, dict):
                self.config = CommentedMap(config.items())
            else:
                self.config = config
            if not isinstance(self.config, CommentedMap):
                raise UnfurlValidationError("invalid YAML document: %s" % self.config)

            findAnchor(self.config, None)  # create _anchorCache
            self._cachedDocIncludes = {}
            # schema should include defaults but can't validate because it doesn't understand includes
            # but should work most of time
            self.config.loadTemplate = self.loadInclude
            self.loadHook = loadHook

            self.baseDirs = [self.getBaseDir()]
            self.includes, expandedConfig = expandDoc(
                self.config, cls=makeMapWithBase(self.config, self.baseDirs[0])
            )
            self.expanded = expandedConfig
            # print('expanded')
            # yaml.dump(config, sys.stdout)
            errors = schema and self.validate(expandedConfig)
            if errors and validate:
                (message, schemaErrors) = errors
                raise UnfurlValidationError(
                    "JSON Schema validation failed: " + message, errors
                )
            else:
                self.valid = not not errors
        except:
            if self.path:
                msg = "Unable to load yaml config at %s" % self.path
            else:
                msg = "Unable to parse yaml config"
            raise UnfurlError(msg, True)

    def loadYaml(self, path, baseDir=None, warnWhenNotFound=False):
        path = os.path.abspath(os.path.join(baseDir or self.getBaseDir(), path))
        if warnWhenNotFound and not os.path.isfile(path):
            return path, None
        with open(path, "r") as f:
            config = self.yaml.load(f)
        return path, config

    def getBaseDir(self):
        if self.path:
            return os.path.dirname(self.path)
        else:
            return "."

    def restoreIncludes(self, changed):
        restoreIncludes(self.includes, self.config, changed, cls=CommentedMap)

    def dump(self, out=sys.stdout):
        self.yaml.dump(self.config, out)

    @property
    def yaml(self):
        if not self._yaml:
            self._yaml = makeYAML(self.vault)
        return self._yaml

    def __getstate__(self):
        state = self.__dict__.copy()
        state["_yaml"] = None
        return state

    def validate(self, config):
        if isinstance(self.schema, six.string_types):
            # assume its a file path
            path = self.schema
            with open(path) as fp:
                self.schema = json.load(fp)
        else:
            path = None
        baseUri = None
        if path:
            baseUri = urljoin("file:", pathname2url(path))
        return findSchemaErrors(config, self.schema, baseUri)

    def loadInclude(self, templatePath, warnWhenNotFound=False):
        if templatePath == self.baseDirs[-1]:
            self.baseDirs.pop()
            return

        if isinstance(templatePath, dict):
            value = templatePath.get("merge")
            if "file" not in templatePath:
                raise UnfurlError(
                    "`file` key missing from document include: %s" % templatePath
                )
            key = templatePath["file"]
        else:
            value = None
            key = templatePath

        if key in self._cachedDocIncludes:
            path, template = self._cachedDocIncludes[key]
            baseDir = os.path.dirname(path)
            self.baseDirs.append(baseDir)
            return value, template, baseDir

        try:
            baseDir = self.baseDirs[-1]
            if self.loadHook:
                path, template = self.loadHook(
                    self, templatePath, baseDir, warnWhenNotFound
                )
            else:
                path, template = self.loadYaml(key, baseDir, warnWhenNotFound)
            newBaseDir = os.path.dirname(path)
            if template is None:
                if warnWhenNotFound:
                    logger.warning(
                        "document include %s does not exist (base: %s)"
                        % (templatePath, baseDir)
                    )
                return value, template, newBaseDir
            elif isinstance(template, CommentedMap):
                template.baseDir = newBaseDir
            _cacheAnchors(self.config._anchorCache, template)
        except:
            raise UnfurlError(
                "unable to load document include: %s (base: %s)"
                % (templatePath, baseDir),
                True,
                True,
            )
        self.baseDirs.append(newBaseDir)

        self._cachedDocIncludes[key] = [path, template]
        return value, template, newBaseDir



import unittest
import os
import traceback
import six
from collections import MutableSequence
from click.testing import CliRunner
from unfurl.__main__ import cli
import unfurl
from unfurl.configurator import Configurator
from unfurl.localenv import LocalEnv
from unfurl.util import sensitive_list
from unfurl.yamlloader import yaml

# python 2.7 needs these:
from unfurl.configurators.shell import ShellConfigurator
from unfurl.configurators.ansible import AnsibleConfigurator

manifest = """
apiVersion: unfurl/v1alpha1
kind: Ensemble
context:
  locals:
    properties:
      prop2:
       type: number
spec:
  node_templates:
    test:
      type: tosca.nodes.Root
      properties:
        local1:
          eval:
            local: prop1
        local2:
          eval:
            external: local
          foreach: prop2
        testApikey:
          eval:
           secret: testApikey
        # uses jinja2 native types to evaluate to a list
        aListOfItems:
            eval:
              template: >
                [{% for key, value in aLocalDict.items() %}
                  {
                    'key': '{{ key }}',
                    'value': '{{ value | b64encode }}'
                  },
                {% endfor %}]
            vars:
              aLocalDict:
                eval:
                  secret: aDict
            # trace: 1
      interfaces:
        Standard:
          create: test_cli.CliTestConfigurator
"""

localConfig = """
apiVersion: unfurl/v1alpha1
kind: Project
contexts:
  defaults: #used if manifest isnt found in `manifests` list below
   secrets:
    attributes:
      aDict:
        key1: a string
        key2: 2
      default: # if key isn't found, apply this:
        eval:
          lookup:
            env: "UNFURL_{{ key | upper }}"
  test:
    locals:
      attributes:
        prop1: 'found'
        prop2: 1
        aDict:
          key1: a string
          key2: 2

manifests:
  - file: git/default-manifest.yaml
    context: test
"""


class CliTestConfigurator(Configurator):
    def run(self, task):
        attrs = task.target.attributes
        assert isinstance(attrs["aListOfItems"], MutableSequence), type(
            attrs["aListOfItems"]
        )
        # sort for python2
        assert sorted(attrs["aListOfItems"], key=lambda k: k["key"]) == [
            {"key": "key1", "value": "YSBzdHJpbmc="},
            {"key": "key2", "value": "Mg=="},
        ], attrs["aListOfItems"]
        assert attrs["local1"] == "found", attrs["local1"]
        assert attrs["local2"] == 1, attrs["local2"]
        assert (
            attrs["testApikey"] == "secret"
        ), "failed to get secret environment variable, maybe DelegateAttributes is broken?"

        # list will be marked as sensitive because the template that created referenced a sensitive content
        assert isinstance(attrs["aListOfItems"], sensitive_list), type(
            attrs["aListOfItems"]
        )
        out = six.StringIO()
        yaml.dump(attrs["aListOfItems"], out)
        assert out.getvalue() == "<<REDACTED>>\n...\n", repr(out.getvalue())
        assert isinstance(
            attrs._attributes["aListOfItems"].asRef(), sensitive_list
        ), type(attrs._attributes["aListOfItems"].asRef())

        yield task.done(True, False)


class CliTest(unittest.TestCase):
    def test_help(self):
        runner = CliRunner()
        result = runner.invoke(cli, [])
        assert result.output.startswith(
            "Usage: cli [OPTIONS] COMMAND [ARGS]"
        ), result.output
        self.assertEqual(result.exit_code, 0)

    def test_version(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["version"])
        self.assertEqual(result.exit_code, 0, result)
        self.assertIn(
            unfurl.__version__(True), result.output.strip()
        )

    def test_badargs(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["--badarg"])
        self.assertEqual(result.exit_code, 2, result)

    def test_run(self):
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open("ensemble.yaml", "w") as f:
                f.write("invalid manifest")

            result = runner.invoke(cli, ["run"])
            self.assertEqual(result.exit_code, 1, result)
            # XXX log handler is writing to the CliRunner's output stream
            self.assertIn("Unable to create job", result.output.strip())

            result = runner.invoke(cli, ["run", "--ensemble", "missing.yaml"])
            assert "Ensemble manifest does not exist" in str(
                result.exception
            ), result.exception

            with open("manifest2.yaml", "w") as f:
                f.write(manifest)
            runCmd = ["run", "--ensemble", "manifest2.yaml"]
            result = runner.invoke(cli, runCmd + ["--", "echo", "ok"])
            # print("result.output1!", result.output)
            assert not result.exception, "\n".join(
                traceback.format_exception(*result.exc_info)
            )
            assert r"'stdout': 'ok\n'" in result.output.replace(
                "u'", "'"
            ), result.output
            # run same command using ansible
            result = runner.invoke(
                cli, runCmd + ["--host", "localhost", "--", "echo", "ok"]
            )
            assert not result.exception, "\n".join(
                traceback.format_exception(*result.exc_info)
            )
            assert r"'stdout': 'ok'" in result.output.replace("u'", "'"), result.output

    def test_localConfig(self):
        # test loading the default manifest declared in the local config
        # test locals and secrets:
        #    declared attributes and default lookup
        #    inherited from default (inheritFrom)
        #    verify secret contents isn't saved in config
        os.environ["UNFURL_TESTAPIKEY"] = "secret"
        runner = CliRunner()
        with runner.isolated_filesystem() as tempDir:
            with open("unfurl.yaml", "w") as local:
                local.write(localConfig)
            repoDir = "git"
            os.mkdir(repoDir)
            os.chdir(repoDir)
            with open("default-manifest.yaml", "w") as f:
                f.write(manifest)

            # make sure the test environment set UNFURL_HOME:
            testHomePath = os.environ["UNFURL_HOME"]
            assert testHomePath and testHomePath.endswith(
                "/tmp/unfurl_home"
            ), testHomePath

            # empty UNFURL_HOME disables the home path
            os.environ["UNFURL_HOME"] = ""
            homePath = LocalEnv().homeConfigPath
            assert not homePath, homePath

            # no UNFURL_HOME and the default home path will be used
            del os.environ["UNFURL_HOME"]
            # we don't want to want to try to load the real home path so call getHomeConfigPath() instead
            # homePath = LocalEnv().homeConfigPath
            homePath = unfurl.getHomeConfigPath(None)
            assert homePath and homePath.endswith(".unfurl_home/unfurl.yaml"), homePath

            # restore test environment's UNFURL_HOME
            os.environ["UNFURL_HOME"] = testHomePath
            homePath = LocalEnv().homeConfigPath
            assert homePath and homePath.startswith(testHomePath), homePath

            homePath = LocalEnv(homePath="").homeConfigPath
            assert homePath is None, homePath

            homePath = LocalEnv(homePath="override").homeConfigPath
            assert homePath and homePath.endswith("/override/unfurl.yaml"), homePath

            result = runner.invoke(cli, ["deploy", "--jobexitcode", "degraded"])
            # uncomment this to see output:
            # print("result.output", result.exit_code, result.output)
            assert not result.exception, "\n".join(
                traceback.format_exception(*result.exc_info)
            )
            self.assertEqual(result.exit_code, 0, result.output)

    def test_clone(self):
        runner = CliRunner()
        with runner.isolated_filesystem():
            specTemplate = os.path.join(
                os.path.dirname(__file__), "examples/spec/service-template.yaml"
            )
            result = runner.invoke(cli, ["clone", specTemplate, "clone1"])
            # print("result.output", result.exit_code, result.output)
            assert not result.exception, "\n".join(
                traceback.format_exception(*result.exc_info)
            )
            self.assertEqual(result.exit_code, 0, result)

            result = runner.invoke(cli, ["deploy", "clone1"])
            # print("result.output", result.exit_code, result.output)
            assert not result.exception, "\n".join(
                traceback.format_exception(*result.exc_info)
            )
            getattr(self, "assertRegex", self.assertRegexpMatches)(
                result.output,
                "Job A[0-9A-Za-z]{11} completed: ok. Found nothing to do.",
            )
            self.assertEqual(result.exit_code, 0, result.output)

            # this will clone the new ensemble
            result = runner.invoke(cli, ["clone", "clone1", "clone2"])
            # print("result.output", result.exit_code, result.output)
            assert not result.exception, "\n".join(
                traceback.format_exception(*result.exc_info)
            )
            self.assertEqual(result.exit_code, 0, result)

            result = runner.invoke(cli, ["deploy", "clone2"])
            # print("result.output", result.exit_code, result.output)
            assert not result.exception, "\n".join(
                traceback.format_exception(*result.exc_info)
            )
            getattr(self, "assertRegex", self.assertRegexpMatches)(
                result.output,
                "Job A[0-9A-Za-z]{11} completed: ok. Found nothing to do.",
            )
            self.assertEqual(result.exit_code, 0, result.output)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import logging

from toscaparser.common.exception import ExceptionCollector
from toscaparser.common.exception import UnknownFieldError, ValidationError
from toscaparser.entity_template import EntityTemplate
from toscaparser.steps import Step
from toscaparser.utils import validateutils
from toscaparser.activities import Precondition

SECTIONS = (METADATA, DESCRIPTION, INPUTS, PRECONDITIONS, STEPS, IMPLEMENTATION, OUTPUTS) = \
           ('metadata', 'description',
            'inputs', 'preconditions', 'steps', 'implementation', 'outputs')

log = logging.getLogger('tosca')


class Workflow(object):
    '''Workflows defined in Topology template.'''
    def __init__(self, name, workflow, custom_def=None):
        self.name = name
        self._tpl = workflow
        self.meta_data = None
        if METADATA in workflow:
            self.meta_data = workflow.get(METADATA)
            validateutils.validate_map(self.meta_data)
        self._validate_keys()
        self.steps = self._steps(workflow.get(STEPS))
        self._validate_steps()
        self.inputs = workflow.get('inputs')
        self.preconditions = [Precondition(tpl) for tpl in workflow.get(PRECONDITIONS, [])]
        self.outputs = workflow.get('outputs')

    def _validate_steps(self):
      steps = set()
      for step in self.steps.values():
        steps.update(step.on_success + step.on_failure)
      missing = steps.difference(self.steps)
      if missing:
          msg = "Workflow %s referencing missing step(s): %s", (self.name, missing)
          ExceptionCollector.appendException(
              ValidationError(message=msg))

    @property
    def description(self):
        return self._tpl.get('description')

    @property
    def metadata(self):
        return self._tpl.get('metadata')

    def _steps(self, steps):
        return {name: Step(name, step_tpl) for name, step_tpl in steps.items()}

    def _validate_keys(self):
        for key in self._tpl.keys():
            if key not in SECTIONS:
                ExceptionCollector.appendException(
                    UnknownFieldError(what='Workflow "%s"' % self.name,
                                      field=key))



"""
This module defines the core model and implements the runtime operations of the model.

The state of the system is represented as a collection of Resources.
Each resource have a status; attributes that describe its state; and a TOSCA template
which describes its capabilities, relationships and available interfaces for configuring and interacting with it.
"""
import six
import collections

from ansible.parsing.dataloader import DataLoader

from .util import UnfurlError, loadClass, toEnum, makeTempDir, ChainMap
from .result import ResourceRef, ChangeAware

from .support import AttributeManager, Defaults, Status, Priority, NodeState, Templar
from .tosca import CapabilitySpec, RelationshipSpec, NodeSpec, TopologySpec

import logging

logger = logging.getLogger("unfurl")

# CF 3.4.1 Node States p61


class Operational(ChangeAware):
    """
  This is an abstract base class for Jobs, Resources, and Configurations all have a Status associated with them
  and all use the same algorithm to compute their status from their dependent resouces, tasks, and configurations
  """

    # XXX3 add repairable, messages?

    # core properties to override
    @property
    def priority(self):
        return Defaults.shouldRun

    @property
    def localStatus(self):
        return Status.unknown

    @property
    def state(self):
        return NodeState.initial

    def getOperationalDependencies(self):
        return ()

    @property
    def manualOverideStatus(self):
        return None

    # derived properties:
    @property
    def operational(self):
        return self.status == Status.ok or self.status == Status.degraded

    @property
    def active(self):
        return self.status <= Status.error

    @property
    def present(self):
        return self.operational or self.status == Status.error

    @property
    def missing(self):
        return self.status == Status.pending or self.status == Status.absent

    @property
    def status(self):
        """
        Return the effective status, considering first the local readyState and
        then the aggregate status' of its dependencies (see `aggregateStatus()`
        and `getOperationalDependencies()`).

        If the localStatus is non-operational, that takes precedence.
        Otherwise the localStatus is compared with its aggregate dependent status
        and worser value is choosen.
        """
        if self.manualOverideStatus is not None:
            status = self.manualOverideStatus
            if status >= Status.error:
                return status
        else:
            status = self.localStatus

        if not status:
            return Status.unknown

        if status >= Status.error:
            # return error, pending, or absent
            return status

        dependentStatus = self.aggregateStatus(self.getOperationalDependencies())
        if dependentStatus is None:
            # note if localStatus is a no op (status purely determined by dependents)
            # it should be set to ok
            # then ok + None = ok, which makes sense, since it has no dependendents
            return status  # return local status
        else:
            # local status is ok, degraded
            return max(status, dependentStatus)

    @property
    def required(self):
        return self.priority == Priority.required

    # `lastConfigChange` indicated the last configuration change applied to the instance.
    # `lastStateChange` is last observed change to the internal state of the instance. This may be reflected in attributes of the instance or completely opaque to the service template's model.

    # The computed attribute `lastChange` is whichever of the above two attributes are more recent.

    @property
    def lastStateChange(self):
        return None

    @property
    def lastConfigChange(self):
        return None

    @property
    def lastChange(self):
        if not self.lastStateChange:
            return self.lastConfigChange
        elif not self.lastConfigChange:
            return self.lastStateChange
        else:
            return max(self.lastStateChange, self.lastConfigChange)

    def hasChanged(self, changeset):
        # if changed since the last time we checked
        if not self.lastChange:
            return False
        if not changeset:
            return True
        return self.lastChange > changeset.changeId

    @staticmethod
    def aggregateStatus(statuses):
        """
        Returns: ok, degraded, pending or None

        If there are no instances, return None
        If any required are not operational, return pending or error
        If any other are not operational or degraded, return degraded
        Otherwise return ok.
        (Instances with priority set to "ignore" are ignored.)
        """
        # error if a configuration is required and not operational
        # error if a non-configuration managed child resource is required and not operational
        # degraded if non-required configurations and resources are not operational
        #          or required configurations and resources are degraded
        # ok otherwise

        # any required is pending then aggregate is pending
        # otherwise just set to degraded
        aggregate = None
        for status in statuses:
            assert isinstance(status, Operational), status
            if aggregate is None:
                aggregate = Status.ok
            if status.priority == Priority.ignore:
                continue
            elif status.required and not status.operational:
                if status.status == Status.pending:
                    aggregate = Status.pending
                else:
                    aggregate = Status.error
                    break
            else:
                if aggregate <= Status.degraded:
                    if not status.operational or status.status == Status.degraded:
                        aggregate = Status.degraded
        return aggregate


class OperationalInstance(Operational):
    def __init__(
        self,
        status=None,
        priority=None,
        manualOveride=None,
        lastStateChange=None,
        lastConfigChange=None,
        state=None,
    ):
        if isinstance(status, OperationalInstance):
            self._localStatus = status._localStatus
            self._manualOverideStatus = status._manualOverideStatus
            self._priority = status._priority
            self._lastStateChange = status._lastStateChange
            self._lastConfigChange = status._lastConfigChange
            self._state = status._state
            self.dependencies = status.dependencies
        else:
            self._localStatus = toEnum(Status, status)
            self._manualOverideStatus = toEnum(Status, manualOveride)
            self._priority = toEnum(Priority, priority)
            self._lastStateChange = lastStateChange
            self._lastConfigChange = lastConfigChange
            self._state = state
            self.dependencies = []
        # self.repairable = False # XXX3
        # self.messages = [] # XXX3

    def getOperationalDependencies(self):
        return self.dependencies

    def localStatus():
        doc = "The localStatus property."

        def fget(self):
            return self._localStatus

        def fset(self, value):
            self._localStatus = value

        def fdel(self):
            del self._localStatus

        return locals()

    localStatus = property(**localStatus())

    def manualOverideStatus():
        doc = "The manualOverideStatus property."

        def fget(self):
            return self._manualOverideStatus

        def fset(self, value):
            self._manualOverideStatus = value

        def fdel(self):
            del self._manualOverideStatus

        return locals()

    manualOverideStatus = property(**manualOverideStatus())

    def priority():
        doc = "The priority property."

        def fget(self):
            return Defaults.shouldRun if self._priority is None else self._priority

        def fset(self, value):
            self._priority = value

        def fdel(self):
            del self._priority

        return locals()

    priority = property(**priority())

    @property
    def lastStateChange(self):
        return self._lastStateChange

    @property
    def lastConfigChange(self):
        return self._lastConfigChange

    def state():
        doc = "The state property."

        def fget(self):
            return self._state

        def fset(self, value):
            self._state = toEnum(NodeState, value)

        return locals()

    state = property(**state())


class _ChildResources(collections.Mapping):
    def __init__(self, resource):
        self.resource = resource

    def __getitem__(self, key):
        return self.resource.findResource(key)

    def __iter__(self):
        return iter(r.name for r in self.resource.getSelfAndDescendents())

    def __len__(self):
        return len(tuple(self.resource.getSelfAndDescendents()))


class EntityInstance(OperationalInstance, ResourceRef):
    attributeManager = None
    created = None
    shadow = None
    imports = None
    envRules = None
    _baseDir = ""

    def __init__(
        self, name="", attributes=None, parent=None, template=None, status=None
    ):
        OperationalInstance.__init__(self, status)
        self.name = name
        self._attributes = attributes or {}
        self.parent = parent
        if parent:
            getattr(parent, self.parentRelation).append(self)

        self.template = template or self.templateType()

    def _resolve(self, key):
        # might return a Result
        self.attributes[key]  # force resolve
        return self.attributes._attributes[key]

    def query(self, expr, vars=None, wantList=False):
        from .eval import Ref, RefContext

        return Ref(expr).resolve(RefContext(self, vars=vars), wantList)

    def localStatus():
        doc = "The localStatus property."

        def fget(self):
            return self._localStatus

        def fset(self, value):
            if self.root.attributeManager:
                self.root.attributeManager.setStatus(self, value)
            self._localStatus = value

        def fdel(self):
            del self._localStatus

        return locals()

    localStatus = property(**localStatus())

    def getOperationalDependencies(self):
        if self.parent and self.parent is not self.root:
            yield self.parent

    @property
    def key(self):
        return "%s::.%s::%s" % (self.parent.key, self.parentRelation[1:], self.name)

    def asRef(self, options=None):
        return {"ref": self.key}

    @property
    def tosca_id(self):
        return self.key

    @property
    def tosca_name(self):
        return self.template.name

    @property
    def type(self):
        return self.template.type

    @property
    def baseDir(self):
        if self.shadow:
            return self.shadow.baseDir
        else:
            return self.root._baseDir

    @property
    def artifacts(self):
        return self.template.artifacts

    @property
    def attributes(self):
        """attributes are live values but _attributes will be the serialized value
    """
        if not self.root.attributeManager:
            if not self.attributeManager:
                # inefficient but create a local one for now
                self.attributeManager = AttributeManager()
            # XXX3 changes to self.attributes aren't saved
            return self.attributeManager.getAttributes(self)

        # returned registered attribute or create a new one
        # attribute class getter resolves references
        return self.root.attributeManager.getAttributes(self)

    @property
    def names(self):
        return self.attributes

    def __eq__(self, other):
        if self is other:
            return True
        if self.__class__ != other.__class__:
            return False
        if self.root is not other.root:
            if self.shadow and self.shadow == other:
                return True
            if other.shadow and other.shadow == self:
                return True
            else:
                return False
        if not self.lastChange:
            # only support equality if resource has a changeid
            return False
        return self.lastChange == other.lastChange and self.key == other.key

    def __getstate__(self):
        state = self.__dict__.copy()
        # Remove the unpicklable entries.
        if state.get("_templar"):
            del state["_templar"]
        if state.get("_interfaces"):
            state["_interfaces"] = {}
        if "attributeManager" in state:
            del state["attributeManager"]
        return state

    def __repr__(self):
        return "%s('%s')" % (self.__class__, self.name)


# both have occurrences
# only need to configure capabilities as required by a relationship
class CapabilityInstance(EntityInstance):
    # 3.7.2 Capability definition p. 97
    # 3.8.1 Capability assignment p. 114
    parentRelation = "_capabilities"
    templateType = CapabilitySpec
    _relationships = None

    @property
    def relationships(self):
        # create relationship instances from the corresponding relationship templates
        if self._relationships is None:
            self._relationships = []
            for template in self.template.relationships:
                # template will be a RelationshipSpec
                assert template.source
                # the constructor will add itself to _relationships
                rel = RelationshipInstance(
                    template.name, parent=self, template=template
                )
                # find the node instance that uses this requirement
                sourceNode = self.root.findResource(template.source.name)
                if sourceNode:
                    rel.source = sourceNode

        return self._relationships

    @property
    def key(self):
        # XXX implement something like _ChildResources to enable ::name instead of [.name]
        return "%s::.%s::[.name=%s]" % (self.parent.key, "capabilities", self.name)


class RelationshipInstance(EntityInstance):
    # 3.7.3 Requirements definition p. 99
    # 3.8.2 Requirements assignment p. 115
    parentRelation = "_relationships"
    templateType = RelationshipSpec
    source = None

    def __init__(
        self, name="", attributes=None, parent=None, template=None, status=None
    ):
        EntityInstance.__init__(self, name, attributes, parent, template, status)

    @property
    def target(self):
        # parent is a capability, return it's parent (a Node)
        return self.parent.parent if self.parent else None

    @property
    def key(self):
        # XXX implement something like _ChildResources to enable ::name instead of [.name=name]
        if self.source:
            return "%s::.requirements::[.name=%s]" % (self.source.key, self.name)
        elif self.parent is self.root:  # it's a default relationship
            return "::.requirements::[.name=%s]" % self.name
        else:  # capability is parent
            return "%s::.relationships::[.name=%s]" % (self.parent.key, self.name)

    def mergeProps(self, matchfn):
        env = {}
        capability = self.parent
        for name, val in capability.template.findProps(capability.attributes, matchfn):
            if val is not None:
                env[name] = val
        for name, val in self.template.findProps(self.attributes, matchfn):
            if val is not None:
                env[name] = val
        return env


class NodeInstance(EntityInstance):
    templateType = NodeSpec
    parentRelation = "instances"

    def __init__(
        self, name="", attributes=None, parent=None, template=None, status=None
    ):
        if parent:
            # only node instances have unique names
            if parent.root.findResource(name):
                raise UnfurlError(
                    'can not create node instance "%s", its name is already in use'
                    % name
                )

        # next two maybe initialized either by Manifest.createNodeInstance or by its property
        self._capabilities = None
        self._requirements = None
        self.instances = []
        EntityInstance.__init__(self, name, attributes, parent, template, status)

        if self.root is self:
            self._all = _ChildResources(self)
            self._templar = Templar(DataLoader())

        self._interfaces = {}
        # preload
        self.getInterface("inherit")
        self.getInterface("default")

    @property
    def requirements(self):
        if self._requirements is None:

            def findRelationship(name, template):
                assert template.relationship  # template is a RequirementSpec
                relationship = template.relationship
                assert relationship and relationship.capability and relationship.target
                # find the Capability instance that corresponds to this relationship template's capability
                targetNodeInstance = self.root.findResource(relationship.target.name)
                assert targetNodeInstance, (
                    "target instance %s should have been already created"
                    % relationship.target.name
                )
                # calling getCapabilities() will create any needed capability instances
                capabilities = targetNodeInstance.getCapabilities(
                    relationship.capability.name
                )
                if capabilities:  # they should have already been created
                    # invoking the relationships property will create any needed relationship instances
                    for relInstance in capabilities[0].relationships:
                        if relInstance.source:
                            if relInstance.source is self:
                                return relInstance
                        else:
                            if relInstance.template.source == self.template:
                                # this instance didn't exist when relInstance was created, assign source now
                                relInstance.source = self
                                return relInstance

                raise UnfurlError(
                    'can not find relation instance for requirement "%s" on node "%s"'
                    % (name, self.name)
                )

            self._requirements = [
                findRelationship(k, v) for (k, v) in self.template.requirements.items()
            ]

        return self._requirements

    def getRequirements(self, match):
        if isinstance(match, six.string_types):
            return [r for r in self.requirements if r.template.name == match]
        elif isinstance(match, NodeInstance):
            return [r for r in self.requirements if r.target == match]
        elif isinstance(match, CapabilityInstance):
            return [r for r in self.requirements if r.parent == match]
        else:
            raise UnfurlError('invalid match for getRequirements: "%s"' % match)

    @property
    def capabilities(self):
        if self._capabilities is None:
            self._capabilities = []
            for name, template in self.template.capabilities.items():
                # if capabilities:
                #    # append index to name if multiple Capabilities exist for that name
                #    name += str(len(capabilities))
                # the new instance will be added to _capabilities
                CapabilityInstance(template.name, parent=self, template=template)

        return self._capabilities

    def getCapabilities(self, name):
        return [
            capability
            for capability in self.capabilities
            if capability.template.name == name
        ]

    def _getDefaultRelationships(self, relation=None):
        if self.root is self:
            return
        for rel in self.root.getDefaultRelationships(relation):
            for capability in self.capabilities:
                if rel.template.matches_target(capability.template):
                    yield rel

    def getDefaultRelationships(self, relation=None):
        return list(self._getDefaultRelationships(relation))

    @property
    def names(self):
        # matches TOSCA logic for get_attribute lookup
        return ChainMap(
            self.attributes,
            {r.name: r.parent for r in self.requirements},
            {c.name: c for c in self.capabilities},
        )

    def _resolve(self, key):
        # might return a Result
        try:
            self.attributes[key]  # force resolve
            return self.attributes._attributes[key]
        except KeyError:
            try:
                inherit = self._interfaces.get("inherit")  # pre-loaded
                if inherit:
                    return inherit(key)
                else:
                    raise
            except KeyError:
                default = self._interfaces.get("default")  # pre-loaded
                if default:
                    return default(key)
                else:
                    raise

    @property
    def key(self):
        return "::%s" % self.name

    def getOperationalDependencies(self):
        # XXX
        # for instance in self.requirements:
        #  yield instance
        # else:
        if self.parent and self.parent is not self.root:
            yield self.parent

    def getSelfAndDescendents(self):
        "Recursive descendent including self"
        yield self
        for r in self.instances:
            for descendent in r.getSelfAndDescendents():
                yield descendent

    @property
    def descendents(self):
        return list(self.getSelfAndDescendents())

    def findResource(self, resourceid):
        if self.name == resourceid:
            return self
        for r in self.instances:
            child = r.findResource(resourceid)
            if child:
                return child
        return None

    def findInstanceOrExternal(self, resourceid):
        instance = self.findResource(resourceid)
        if instance:
            return instance
        if self.imports:
            return self.imports.findImport(resourceid)
        return None

    def addInterface(self, klass, name=None):
        if not isinstance(klass, six.string_types):
            klass = klass.__module__ + "." + klass.__name__
        current = self._attributes.setdefault(".interfaces", {})
        current[name or klass] = klass
        return current

    def getInterface(self, name):
        # XXX uses TOSCA interfaces instead
        if ".interfaces" not in self._attributes:
            return None  # no interfaces

        if not isinstance(name, six.string_types):
            name = name.__module__ + "." + name.__name__
        instance = self._interfaces.get(name)
        if instance:
            return instance
        else:
            className = self._attributes[".interfaces"].get(name)
            if className:
                instance = loadClass(className)(name, self)
                self._interfaces[name] = instance
                return instance
        return None

    def __repr__(self):
        return "NodeInstance('%s')" % self.name


class TopologyInstance(NodeInstance):
    templateType = TopologySpec

    def __init__(self, template, status=None):
        NodeInstance.__init__(self, "root", template=template, status=status)
        # add these as special child resources so they can be accessed like "::inputs::foo"
        self.inputs = NodeInstance("inputs", template.inputs, self)
        self.outputs = NodeInstance("outputs", template.outputs, self)
        self._capabilities = []
        self._relationships = None
        self._tmpDir = None

    def setBaseDir(self, baseDir):
        self._baseDir = baseDir
        if not self._templar or self._templar._basedir != baseDir:
            loader = DataLoader()
            loader.set_basedir(baseDir)
            self._templar = Templar(loader)

    @property
    def requirements(self):
        """
        The root node returns RelationshipInstances representing default relationship templates
        """
        if self._relationships is None:
            self._relationships = []
            relTemplates = self.template.spec.relationshipTemplates
            for name, template in relTemplates.items():
                # template will be a RelationshipSpec
                if template.toscaEntityTemplate.default_for:
                    # the constructor will add itself to _relationships
                    # note: name might not equal template.name if from an external spec
                    RelationshipInstance(name, parent=self, template=template)

        return self._relationships

    def getDefaultRelationships(self, relation=None):
        # for root, this is the same as self.requirements
        if not relation:
            return self.requirements
        return [
            rel for rel in self.requirements if rel.template.isCompatibleType(relation)
        ]

    def getOperationalDependencies(self):
        for instance in self.instances:
            if instance.name not in ["inputs", "outputs"]:
                yield instance

    @property
    def tmpDir(self):
        if not self._tmpDir:
            self._tmpDir = makeTempDir()
        return self._tmpDir



from __future__ import absolute_import
import sys
import collections
import functools
import logging
from ..util import assertForm, saveToTempfile
from ..configurator import Status
from ..result import serializeValue
from . import TemplateConfigurator
import ansible.constants as C
from ansible import context
from ansible.cli.playbook import PlaybookCLI
from ansible.plugins.callback.default import CallbackModule
from ansible.module_utils import six
from ansible.utils.display import Display

display = Display()

logger = logging.getLogger("unfurl")

# inputs:
#  playbook
#  playbookArgs
#  extraVars
#  inventory
# outputs:
#  treated as ansible facts to extract from ansible results
def getAnsibleResults(result, extraKeys=(), facts=()):
    """
  Returns a dictionary containing at least:

  msg
  stdout
  returncode: (None if the process didn't complete)
  error: stderr or exception if one was raised
  **extraKeys
  """
    # result is per-task ansible.executor.task_result.TaskResult
    # https://github.com/ansible/ansible/blob/devel/lib/ansible/executor/task_result.py
    # https://docs.ansible.com/ansible/latest/reference_appendices/common_return_values.html
    # https://docs.ansible.com/ansible/latest/user_guide/playbooks_variables.html#variables-discovered-from-systems-facts

    # _check_key checks 'results' if task was a loop
    # 'warnings': result._check_key('warning'),
    result = result.clean_copy()
    logger.debug("playbook task result: %s", str(result._result))
    resultDict = {}
    # map keys in results to match the names that ShellConfigurator uses
    keyMap = {
        "returncode": ["returncode"],
        "msg": ["msg"],
        "error": ["exception"],
        "stdout": ["stdout", "module_stdout"],
        "stderr": ["module_stderr"],
    }
    # other common keys: "result", "method", "diff", "invocation"
    for name, keys in keyMap.items():
        for key in keys:
            if key in result._result:
                resultDict[name] = result._result[key]
                break
    for key in extraKeys:
        if key in result._result:
            resultDict[key] = result._result[key]

    outputs = {}
    if facts:
        ansible_facts = result._result.get("ansible_facts")
        if ansible_facts:
            for fact in facts:
                if fact in ansible_facts:
                    outputs[fact] = ansible_facts[fact]
    return resultDict, outputs


class AnsibleConfigurator(TemplateConfigurator):
    """The current resource is the inventory."""

    def __init__(self, configSpec):
        super(AnsibleConfigurator, self).__init__(configSpec)
        self._cleanupRoutines = []

    def canDryRun(self, task):
        return True

    def _makeInventoryFromGroup(self, group):
        hosts = {}
        vars = {}
        children = {}
        for member in group.memberInstances:
            hosts[member.name] = self._getHostVars(member)
        for child in group.memberGroups:
            if child.isCompatibleType("unfurl.groups.AnsibleInventoryGroup"):
                children[child] = self._makeInventoryFromGroup(child)
        for prop, value in group.properties.items():
            vars[prop] = value
        return dict(hosts=hosts, vars=vars, children=children)

    def _getHostVars(self, node):
        # return ansible_connection, ansible_host, ansible_user, ansible_port
        connections = node.getCapabilities("endpoint")
        for connection in connections:
            if connection.template.isCompatibleType(
                "unfurl.capabilities.Endpoint.Ansible"
            ):
                break
        else:
            return {}
        props = connection.attributes
        hostVars = {
            "ansible_" + name: props[name]
            for name in ("port", "host", "connection", "user")
            if name in props
        }
        # ansible_user
        hostVars.update(props.get("hostvars", {}))
        if "ansible_host" not in hostVars and hostVars.get("ip_address"):
            hostVars["ansible_host"] = hostVars["ip_address"]
        return hostVars

    def _updateVars(self, connection, hostVars):
        creds = connection.attributes.get("credential")
        if creds:
            if "user" in creds:
                hostVars["ansible_user"] = creds["user"]
            # e.g token_type is password or private_key_file:
            if "token" in creds:
                hostVars["ansible_" + creds["token_type"]] = creds["token"]
            if "keys" in creds:
                hostVars.update(creds["keys"])
        hostVars.update(connection.attributes.get("hostvars", {}))

    def _makeInventory(self, host, allVars, task):
        if host:
            hostVars = self._getHostVars(host)
            connection = task.findConnection(
                host, "unfurl.relationships.ConnectsTo.Ansible"
            )
            if connection:
                self._updateVars(connection, hostVars)
            hosts = {host.name: hostVars}

            children = {
                group.name: self._makeInventoryFromGroup(group)
                for group in host.template.groups
                if group.isCompatibleType("unfurl.groups.AnsibleInventoryGroup")
            }
        else:
            hostVars = {}
            hosts = {"localhost": hostVars}
            children = {}

        if (
            next(iter(hosts)) == "localhost"
            and "ansible_python_interpreter" not in hostVars
        ):
            # we need to set this in case we are running inside a virtual environment, see:
            # https://docs.ansible.com/ansible/latest/scenario_guides/guide_rax.html#running-from-a-python-virtual-environment-optional
            hostVars["ansible_python_interpreter"] = sys.executable

        # note: allVars is inventory vars shared by all hosts
        return dict(all=dict(hosts=hosts, vars=allVars, children=children))

    def getInventory(self, task):
        inventory = task.inputs.get("inventory")
        if inventory and isinstance(inventory, six.string_types):
            # XXX if user set inventory file we can create a folder to merge them
            # https://allandenot.com/devops/2015/01/16/ansible-with-multiple-inventory-files.html
            return inventory  # assume its a file path

        if not inventory:
            # XXX merge inventory
            # default to localhost if not inventory
            inventory = self._makeInventory(task.operationHost, inventory or {}, task)
        # XXX cache and reuse file
        return saveToTempfile(inventory, "-inventory.yaml").name
        # don't worry about the warnings in log, see:
        # https://github.com/ansible/ansible/issues/33132#issuecomment-346575458
        # https://github.com/ansible/ansible/issues/33132#issuecomment-363908285
        # https://github.com/ansible/ansible/issues/48859

    def _cleanup(self):
        for func in self._cleanupRoutines:
            try:
                func()
            except:
                # XXX: log
                pass
        self._cleanupRoutines = []

    def getVars(self, task):
        vars = task.inputs.context.vars.copy()
        vars["__unfurl"] = task.inputs.context
        return vars

    def _makePlayBook(self, playbook, task):
        assertForm(playbook, collections.MutableSequence)
        # XXX use host group instead of localhost depending on operation_host
        hosts = task.operationHost and task.operationHost.name or "localhost"
        if playbook and not "hosts" in playbook[0]:
            play = dict(hosts=hosts, gather_facts=False, tasks=playbook)
            if hosts == "localhost":
                play["connection"] = "local"
            return [play]
        else:
            return playbook

    def findPlaybook(self, task):
        return task.inputs["playbook"]

    def getPlaybook(self, task):
        playbook = self.findPlaybook(task)
        if isinstance(playbook, six.string_types):
            # assume it's file path
            return playbook
        playbook = self._makePlayBook(playbook, task)
        envvars = task.getEnvironment(True)
        for play in playbook:
            play["environment"] = envvars
        return saveToTempfile(serializeValue(playbook), "-playbook.yml").name

    def getPlaybookArgs(self, task):
        args = task.inputs.get("playbookArgs", [])
        if not isinstance(args, collections.MutableSequence):
            args = [args]
        if task.dryRun:
            args.append("--check")
        if task.configSpec.timeout:
            args.append("--timeout=%s" % task.configSpec.timeout)
        if task.verbose > 0:
            args.append("-" + ("v" * task.verbose))
        return args

    def getResultKeys(self, task, results):
        return []

    def processResult(self, task, result):
        self.processResultTemplate(task, dict(result.result, outputs=result.outputs))
        return result

    def run(self, task):
        try:
            # build host inventory from resource
            inventory = self.getInventory(task)
            playbook = self.getPlaybook(task)

            # build vars from inputs
            extraVars = self.getVars(task)
            if task.operationHost and task.operationHost.templar:
                vault_secrets = task.operationHost.templar._loader._vault.secrets
            else:
                vault_secrets = None
            resultCallback = runPlaybooks(
                [playbook],
                inventory,
                extraVars,
                self.getPlaybookArgs(task),
                vault_secrets,
            )

            if resultCallback.exit_code or len(resultCallback.resultsByStatus.failed):
                status = Status.error
            else:
                # unreachable, failed, skipped
                # XXX degraded vs. error if required?
                status = Status.ok

            logger.debug(
                "ran playbook: status %s changed %s, total %s ",
                status,
                resultCallback.changed,
                len(resultCallback.results),
            )

            if resultCallback.results:
                resultKeys = self.getResultKeys(task, resultCallback.results)
                factKeys = list(task.configSpec.outputs)
                # each task in a playbook will have a corresponding result
                resultList, outputList = zip(
                    *map(
                        lambda result: getAnsibleResults(result, resultKeys, factKeys),
                        resultCallback.results,
                    )
                )
                mergeFn = lambda a, b: a.update(b) or a
                results = functools.reduce(mergeFn, resultList, {})
                outputs = functools.reduce(mergeFn, outputList, {})
            else:
                results, outputs = {}, {}

            targetStatus = None
            if resultCallback.changed > 0:
                modified = True
                if any(
                    r.is_failed() and r.is_changed() for r in resultCallback.results
                ):
                    targetStatus = Status.error
            else:
                modified = False

            result = task.done(
                status == Status.ok and not task._errors,
                modified,
                targetStatus,
                result=results,
                outputs=outputs,
            )
            if (
                (results or outputs)
                and status == Status.ok
                or status == Status.degraded
            ):
                # this can update resources so don't do it on error
                # XXX align with shell configurator behavior
                # (which is called unconditionally and before done())
                result = self.processResult(task, result)
            yield result

        finally:
            self._cleanup()


# see https://github.com/ansible/ansible/blob/d72587084b4c43746cdb13abb262acf920079865/examples/scripts/uptime.py
# and https://github.com/ansible/ansible/blob/devel/lib/ansible/plugins/callback/log_plays.py
_ResultsByStatus = collections.namedtuple(
    "_ResultsByStatus", "ok failed skipped unreachable"
)


class ResultCallback(CallbackModule):
    # NOTE: callbacks will run in separate process
    # see ansible.executor.task_result.TaskResult and ansible.playbook.task.Task

    def __init__(self):
        super(ResultCallback, self).__init__()
        self.results = []
        # named tuple of OrderedDict<task_name:list<result>>
        self.resultsByStatus = _ResultsByStatus(
            *[collections.OrderedDict() for x in range(4)]
        )
        self._load_name = "result"
        self.changed = 0

    def get_option(self, k):
        return False

    def getInfo(self, result):
        host = result._host
        taskname = result.task_name
        fields = result._task_fields.keys()
        keys = result._result.keys()
        return "%s: %s(%s) => %s" % (host, taskname, fields, keys)

    def _addResult(self, status, result):
        self.results.append(result)
        if result._result.get("changed", False):
            self.changed += 1
        # XXX should save by host too
        getattr(self.resultsByStatus, status).setdefault(result.task_name, []).append(
            result
        )

    def v2_runner_on_ok(self, result):
        self._addResult("ok", result)
        # print("ok", self.getInfo(result))
        super(ResultCallback, self).v2_runner_on_ok(result)

    def v2_runner_on_skipped(self, result):
        self._addResult("skipped", result)
        # print("skipped", self.getInfo(result))
        super(ResultCallback, self).v2_runner_on_skipped(result)

    def v2_runner_on_failed(self, result, **kwargs):
        self._addResult("failed", result)
        # print("failed", self.getInfo(result))
        super(ResultCallback, self).v2_runner_on_failed(result, **kwargs)

    def v2_runner_on_unreachable(self, result):
        self._addResult("unreachable", result)
        # print("unreachable", self.getInfo(result))
        super(ResultCallback, self).v2_runner_on_unreachable(result)


def _init_global_context(options):
    # Context _init_global_context is designed to be called only once,
    # setting CLIARGS to an ImmutableDict singleton
    # Since we need to change it, replace the function with one that hacks into CLIARGS internal state
    context.CLIARGS._store = vars(options)


def runPlaybooks(playbooks, _inventory, params=None, args=None, vault_secrets=None):
    inventoryArgs = ["-i", _inventory] if _inventory else []
    args = ["ansible-playbook"] + inventoryArgs + (args or []) + playbooks
    logger.info("running " + " ".join(args))
    cli = PlaybookCLI(args)

    context._init_global_context = _init_global_context
    cli.parse()

    # replace C.DEFAULT_STDOUT_CALLBACK with our own so we have control over logging
    # config/base.yml sets C.DEFAULT_STDOUT_CALLBACK == 'default' (ansible/plugins/callback/default.py)
    # (cli/console.py and cli/adhoc.py sets it to 'minimal' but PlaybookCLI.run() in cli/playbook.py uses the default)
    # see also https://github.com/projectatomic/atomic-host-tests/blob/master/callback_plugins/default.py
    resultsCB = ResultCallback()
    resultsCB.set_options()
    C.DEFAULT_STDOUT_CALLBACK = resultsCB

    _play_prereqs = cli._play_prereqs

    def hook_play_prereqs():
        loader, inventory, variable_manager = _play_prereqs()
        if vault_secrets:
            loader.set_vault_secrets(vault_secrets)
        if params:
            variable_manager._extra_vars.update(params)
        resultsCB.inventoryManager = inventory
        resultsCB.variableManager = variable_manager
        return loader, inventory, variable_manager

    cli._play_prereqs = hook_play_prereqs

    oldVerbosity = display.verbosity
    if logging.getLogger("unfurl.ansible").getEffectiveLevel() <= 10:  # debug
        display.verbosity = 2
    try:
        resultsCB.exit_code = cli.run()
    finally:
        display.verbosity = oldVerbosity
    return resultsCB



import unfurl.util
import unittest
from unfurl.yamlmanifest import YamlManifest
from unfurl.job import Runner, JobOptions
from unfurl.configurators.ansible import runPlaybooks
from unfurl.runtime import Status
import os
import os.path
import warnings
import sys


class AnsibleTest(unittest.TestCase):
    def setUp(self):
        try:
            # Ansible generates tons of ResourceWarnings
            warnings.simplefilter("ignore", ResourceWarning)
        except:
            # python 2.x doesn't have ResourceWarning
            pass
        self.results = {}

    def runPlaybook(self, args=None):
        return runPlaybooks(
            [os.path.join(os.path.dirname(__file__), "examples", "testplaybook.yaml")],
            "localhost,",
            {
                "ansible_connection": "local",
                "extra": 1,
                "ansible_python_interpreter": sys.executable,  # suppress ansible warning
            },
            args,
        )

    def test_runplaybook(self):
        results = self.runPlaybook()
        self.results["runplaybook"] = results
        self.assertEqual(
            "test",
            results.variableManager._nonpersistent_fact_cache["localhost"].get(
                "one_fact"
            ),
        )

        hostfacts = results.variableManager._nonpersistent_fact_cache["localhost"]
        self.assertEqual(hostfacts["one_fact"], "test")
        self.assertEqual(hostfacts["echoresults"]["stdout"], "hello")

    # setting UNFURL_LOGGING can break this test, so skip if it's set
    @unittest.skipIf(
        os.getenv("UNFURL_LOGGING"), "this test requires default log level"
    )
    def test_verbosity(self):
        results = self.runPlaybook()
        # task test-verbosity was skipped
        assert not results.resultsByStatus.ok.get("test-verbosity")
        assert results.resultsByStatus.skipped.get("test-verbosity")
        results = self.runPlaybook(["-vv"])
        # task test-verbosity was ok this time
        assert results.resultsByStatus.ok.get("test-verbosity")
        assert not results.resultsByStatus.skipped.get("test-verbosity")


manifest = """
apiVersion: unfurl/v1alpha1
kind: Manifest
configurations:
  create:
    implementation: unfurl.configurators.ansible.AnsibleConfigurator
    outputs:
      fact1:
      fact2:
    inputs:
      playbook:
        q:
          - set_fact:
              fact1: "{{ '.name' | ref }}"
              fact2: "{{ SELF.testProp }}"
          - name: Hello
            command: echo "{{hostvars['localhost'].ansible_python_interpreter}}"
spec:
  service_template:
    topology_template:
      node_templates:
        test1:
          type: tosca.nodes.Root
          properties:
            testProp: "test"
          interfaces:
            Standard:
              +/configurations:
"""


class AnsibleConfiguratorTest(unittest.TestCase):
    def setUp(self):
        try:
            # Ansible generates tons of ResourceWarnings
            warnings.simplefilter("ignore", ResourceWarning)
        except:
            # python 2.x doesn't have ResourceWarning
            pass

    def test_configurator(self):
        """
    test that runner figures out the proper tasks to run
    """
        runner = Runner(YamlManifest(manifest))
        run1 = runner.run(JobOptions(resource="test1"))
        assert not run1.unexpectedAbort, run1.unexpectedAbort.getStackTrace()
        assert len(run1.workDone) == 1, run1.workDone
        result = list(run1.workDone.values())[0].result
        self.assertEqual(result.outputs, {"fact1": "test1", "fact2": "test"})
        self.assertEqual(result.result, {"stdout": sys.executable})
        assert run1.status == Status.ok, run1.summary()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from toscaparser.capabilities import Capability
from toscaparser.common.exception import ExceptionCollector
from toscaparser.common.exception import MissingRequiredFieldError
from toscaparser.common.exception import UnknownFieldError
from toscaparser.common.exception import ValidationError
from toscaparser.elements.grouptype import GroupType
from toscaparser.elements.interfaces import InterfacesDef, INTERFACE_DEF_RESERVED_WORDS
from toscaparser.elements.interfaces import CONFIGURE, CONFIGURE_SHORTNAME
from toscaparser.elements.interfaces import LIFECYCLE,  LIFECYCLE_SHORTNAME
from toscaparser.elements.nodetype import NodeType
from toscaparser.elements.policytype import PolicyType
from toscaparser.elements.relationshiptype import RelationshipType
from toscaparser.properties import Property
from toscaparser.unsupportedtype import UnsupportedType
from toscaparser.utils.gettextutils import _
from toscaparser.elements.capabilitytype import CapabilityTypeDef
from toscaparser.elements.artifacttype import ArtifactTypeDef

class EntityTemplate(object):
    '''Base class for TOSCA templates.'''

    SECTIONS = (DERIVED_FROM, PROPERTIES, REQUIREMENTS,
                INTERFACES, CAPABILITIES, TYPE, DESCRIPTION, DIRECTIVES, INSTANCE_KEYS,
                ATTRIBUTES, ARTIFACTS, NODE_FILTER, COPY) = \
               ('derived_from', 'properties', 'requirements', 'interfaces',
                'capabilities', 'type', 'description', 'directives', "instance_keys",
                'attributes', 'artifacts', 'node_filter', 'copy')
    REQUIREMENTS_SECTION = (NODE, CAPABILITY, RELATIONSHIP, OCCURRENCES, NODE_FILTER) = \
                           ('node', 'capability', 'relationship',
                            'occurrences', 'node_filter')
    # Special key names
    SPECIAL_SECTIONS = (METADATA) = ('metadata')

    additionalProperties = True
    _source = None

    def __init__(self, name, template, entity_name, custom_def=None):
        self.name = name
        self.entity_tpl = template
        self.custom_def = custom_def
        self._validate_field(self.entity_tpl)
        type = self.entity_tpl.get('type')
        UnsupportedType.validate_type(type)
        self._validate_fields(template)
        if entity_name == 'node_type':
            self.type_definition = NodeType(type, custom_def) \
                if type is not None else None
            self._validate_directives(self.entity_tpl)
        if entity_name == 'relationship_type':
            self.type_definition = RelationshipType(type, custom_def)
        if entity_name == 'policy_type':
            if not type:
                msg = (_('Policy definition of "%(pname)s" must have'
                       ' a "type" ''attribute.') % dict(pname=name))
                ExceptionCollector.appendException(
                    ValidationError(message=msg))
            self.type_definition = PolicyType(type, custom_def)
        if entity_name == 'group_type':
            self.type_definition = GroupType(type, custom_def) \
                if type is not None else None
        if entity_name == 'artifact_type':
            self.type_definition = ArtifactTypeDef(type, custom_def) \
                if type is not None else None

        self._properties = None
        self._interfaces = None
        self._requirements = None
        self._capabilities = None
        if not self.type_definition:
            msg = "no type found %s for %s"  % (entity_name, template)
            ExceptionCollector.appendException(ValidationError(message=msg))
            return
        metadata = self.type_definition.get_definition('metadata')
        if metadata and 'additionalProperties' in metadata:
            self.additionalProperties = metadata['additionalProperties']

        self._validate_properties()
        for prop in self.get_properties_objects():
            prop.validate()
        self._validate_interfaces()

    @property
    def type(self):
        if self.type_definition:
            return self.type_definition.type

    @property
    def parent_type(self):
        if self.type_definition:
            return self.type_definition.parent_type

    @property
    def types(self):
        types = []
        p = self.type_definition
        while p:
            types.append(p)
            p = p.parent_type
        return types

    @property
    def directives(self):
        return self.entity_tpl.get('directives', [])

    @property
    def requirements(self):
        if self._requirements is None:
            self._requirements = self.type_definition.get_value(
                self.REQUIREMENTS,
                self.entity_tpl) or []
        return self._requirements

    def get_properties_objects(self):
        '''Return properties objects for this template.'''
        if self._properties is None:
            self._properties = self._create_properties()
        return self._properties

    def get_properties(self):
        '''Return a dictionary of property name-object pairs.'''
        return {prop.name: prop
                for prop in self.get_properties_objects()}

    def get_property_value(self, name):
        '''Return the value of a given property name.'''
        props = self.get_properties()
        if props and name in props.keys():
            return props[name].value

    @property
    def interfaces(self):
        if self._interfaces is None:
            self._interfaces = self._create_interfaces()
        return self._interfaces

    def get_capabilities_objects(self):
        '''Return capabilities objects for this template.'''
        if not self._capabilities:
            self._capabilities = self._create_capabilities()
        return self._capabilities

    def get_capabilities(self):
        '''Return a dictionary of capability name-object pairs.'''
        return {cap.name: cap
                for cap in self.get_capabilities_objects()}

    def is_derived_from(self, type_str):
        '''Check if object inherits from the given type.

        Returns true if this object is derived from 'type_str'.
        False otherwise.
        '''
        if not self.type:
            return False
        elif self.type == type_str:
            return True
        elif self.parent_type:
            return self.parent_type.is_derived_from(type_str)
        else:
            return False

    def _create_capabilities(self):
        capability = []
        caps = self.type_definition.get_value(self.CAPABILITIES,
                                              self.entity_tpl, parent=True)
        if caps:
            for name, props in caps.items():
                ctype = props.get('type')
                capabilities = self.type_definition.get_capabilities()
                if name in capabilities.keys():
                    c = capabilities[name]
                    ctype = props.get('type')
                    if ctype and ctype != c.type:
                        c = CapabilityTypeDef(name, ctype, self.type_definition.type,
                                                self.type_definition.custom_def)
                    properties = {}
                    # first use the definition default value
                    if c.properties:
                        for property_name in c.properties.keys():
                            prop_def = c.properties[property_name]
                            if 'default' in prop_def:
                                properties[property_name] = prop_def['default']
                    # then update (if available) with the node properties
                    if 'properties' in props and props['properties']:
                        properties.update(props['properties'])

                    cap = Capability(name, properties, c, self.custom_def)
                    capability.append(cap)
        return capability

    def _validate_directives(self, template):
        msg = (_('directives of "%s" must be a list of strings') % self.name)
        keys = template.get("directives", [])
        if not isinstance(keys, list):
            ExceptionCollector.appendException(
                ValidationError(message=msg))
        for key in keys:
            if not isinstance(key, str):
                ExceptionCollector.appendException(
                    ValidationError(message=msg))

    def _validate_properties(self):
        properties = self.type_definition.get_value(self.PROPERTIES, self.entity_tpl)
        self._common_validate_properties(self.type_definition, properties, self.additionalProperties)

    def _validate_capabilities(self):
        type_capabilities = self.type_definition.get_capabilities()
        allowed_caps = \
            type_capabilities.keys() if type_capabilities else []
        capabilities = self.type_definition.get_value(self.CAPABILITIES,
                                                      self.entity_tpl)
        if capabilities:
            self._common_validate_field(capabilities, allowed_caps,
                                        'capabilities')
            self._validate_capabilities_properties(capabilities)

    def _validate_capabilities_properties(self, capabilities):
        for cap, props in capabilities.items():
            capability = self.get_capability(cap)
            if not capability:
                continue
            capabilitydef = capability.type_definition
            self._common_validate_properties(capabilitydef,
                                             props.get(self.PROPERTIES) or {})

            # validating capability properties values
            for prop in self.get_capability(cap).get_properties_objects():
                prop.validate()

                # TODO(srinivas_tadepalli): temporary work around to validate
                # default_instances until standardized in specification
                if cap == "scalable" and prop.name == "default_instances":
                    prop_dict = props[self.PROPERTIES]
                    min_instances = prop_dict.get("min_instances")
                    max_instances = prop_dict.get("max_instances")
                    default_instances = prop_dict.get("default_instances")
                    if not (min_instances <= default_instances
                            <= max_instances):
                        err_msg = ('"properties" of template "%s": '
                                   '"default_instances" value is not between '
                                   '"min_instances" and "max_instances".' %
                                   self.name)
                        ExceptionCollector.appendException(
                            ValidationError(message=err_msg))

    def _common_validate_properties(self, entitytype, properties, allowUndefined=False):
        allowed_props = []
        required_props = []
        for p in entitytype.get_properties_def_objects():
            allowed_props.append(p.name)
            # If property is 'required' and has no 'default' value then record
            if p.required and p.default is None:
                required_props.append(p.name)
        # validate all required properties have values
        if properties:
            req_props_no_value_or_default = []
            if not allowUndefined:
              self._common_validate_field(properties, allowed_props,
                                          'properties')
            # make sure it's not missing any property required by a tosca type
            for r in required_props:
                if r not in properties.keys():
                    req_props_no_value_or_default.append(r)
            # Required properties found without value or a default value
            if req_props_no_value_or_default:
                ExceptionCollector.appendException(
                    MissingRequiredFieldError(
                        what='"properties" of template "%s"' % self.name,
                        required=req_props_no_value_or_default))
        else:
            # Required properties in schema, but not in template
            if required_props:
                ExceptionCollector.appendException(
                    MissingRequiredFieldError(
                        what='"properties" of template "%s"' % self.name,
                        required=required_props))

    def _validate_field(self, template):
        if not isinstance(template, dict):
            ExceptionCollector.appendException(
                MissingRequiredFieldError(
                    what='Template "%s"' % self.name, required=self.TYPE))
        try:
            template[self.TYPE]
        except KeyError:
            ExceptionCollector.appendException(
                MissingRequiredFieldError(
                    what='Template "%s"' % self.name, required=self.TYPE))

    def _common_validate_field(self, schema, allowedlist, section):
        for name in schema:
            if name not in allowedlist:
                ExceptionCollector.appendException(
                    UnknownFieldError(
                        what=('"%(section)s" of template "%(nodename)s"'
                              % {'section': section, 'nodename': self.name}),
                        field=name))

    def _validate_fields(self, template):
        for name in template.keys():
            if name not in self.SECTIONS and name not in self.SPECIAL_SECTIONS:
                ExceptionCollector.appendException(
                    UnknownFieldError(what='template "%s"' % self.name,
                                      field=name))

    def _create_properties(self):
        props = []
        properties = self.type_definition.get_value(self.PROPERTIES,
                                                    self.entity_tpl) or {}
        props_def = self.type_definition.get_properties_def()
        for name, value in properties.items():
            if props_def and name in props_def:
                prop = Property(name, value,
                                props_def[name].schema, self.custom_def)
                props.append(prop)
            elif self.additionalProperties:
                prop = Property(name, value,
                          dict(type='any'), self.custom_def)
                props.append(prop)
        for p in props_def.values():
            if p.default is not None and p.name not in properties:
                prop = Property(p.name, p.default, p.schema, self.custom_def)
                props.append(prop)
        return props

    def _create_interfaces(self):
        interfaces = []
        tpl_interfaces = self.type_definition.get_value(self.INTERFACES,
                                                         self.entity_tpl)

        if self.type_definition.interfaces:
            interfacesDefs = self.type_definition.interfaces.copy()
            if tpl_interfaces:
                # merge the interfaces defined on the type with the template's interface definitions
                for iName, defs in tpl_interfaces.items():
                    # for each interface, see if base defines it too
                    defs = defs.copy()
                    inputs = defs.get('inputs', {})
                    if 'operations' in defs:
                        defs = defs.get('operations', {})
                    baseDefs = interfacesDefs.get(iName)
                    if baseDefs:
                        # add in base's ops and merge interface-level inputs
                        baseInputs = baseDefs.get('inputs')
                        if baseInputs: # merge shared inputs
                            inputs = dict(baseInputs, **inputs)
                            defs['inputs'] = inputs

                        implementation = baseDefs.get('implementation')
                        # set shared implementation
                        if implementation and 'implementation' not in defs:
                            defs['implementation'] = implementation
                            if isinstance(implementation, dict) and self.type_definition._source:
                                # if implementation might be an inline artifact, save the baseDir of the source
                                implementation['_source'] = self.type_definition._source

                        if 'operations' in baseDefs:
                            baseDefs = baseDefs.get('operations') or {}

                        for op, baseDef in baseDefs.items():
                            if op in ['inputs', 'notifications', '_source']:
                                continue
                            if op in defs:
                                # op in both, merge
                                currentiDef = defs[op]
                                if isinstance(baseDef, dict):
                                    if not isinstance(currentiDef, dict):
                                        currentiDef = dict(implementation = currentiDef)
                                    if isinstance(baseDef.get('implementation'), dict) and self.type_definition._source:
                                        # if implementation might be an inline artifact, save the baseDir of the source
                                        baseDef['implementation']['_source'] = self.type_definition._source
                                    defs[op] = dict(baseDef, **currentiDef)
                                    if 'inputs' in baseDef and 'inputs' in currentiDef:
                                        # merge inputs
                                        defs[op]['inputs'] = dict(baseDef['inputs'], **currentiDef['inputs'])
                            else:
                                defs[op] = baseDef

                    # add or replace:
                    interfacesDefs[iName] = defs
        else:
            interfacesDefs = tpl_interfaces
            if not interfacesDefs:
              return []

        defaults = interfacesDefs.pop('defaults', {})
        for interface_type, value in interfacesDefs.items():
            # shared inputs
            inputs = value.get('inputs')
            defaultInputs = defaults.get('inputs')
            if inputs and defaultInputs: # merge shared inputs
                inputs = dict(defaultInputs, **inputs)
            else:
                inputs = inputs or defaultInputs

            # shared outputs
            outputs = value.get('outputs')
            defaultOutputs = defaults.get('outputs')
            if outputs and defaultOutputs: # merge shared inputs
                outputs = dict(defaultOutputs, **outputs)
            else:
                outputs = outputs or defaultOutputs

            # shared implementation
            implementation = value.get('implementation') or defaults.get('implementation')
            _source = value.pop('_source', None)
            if 'operations' in value:
                defs = value.get('operations') or {}
            else:
                defs = value

            for op, op_def in defs.items():
                if op in INTERFACE_DEF_RESERVED_WORDS:
                    continue
                if not isinstance(op_def, dict):
                    op_def = dict(implementation = op_def or implementation)
                elif implementation and not op_def.get('implementation'):
                    op_def['implementation'] = implementation
                if _source:
                    op_def['_source'] = _source
                iface = InterfacesDef(self.type_definition,
                                      interfacetype=interface_type,
                                      node_template=self,
                                      name=op,
                                      value=op_def,
                                      inputs=inputs.copy() if inputs else None,
                                      outputs=outputs.copy() if outputs else None)
                interfaces.append(iface)

            # add a "default" operation that has the shared inputs and implementation
            iface = InterfacesDef(self.type_definition,
                                  interfacetype=interface_type,
                                  node_template=self,
                                  name='default',
                                  value=dict(implementation=implementation,
                                              _source=_source),
                                  inputs=inputs, outputs=outputs)
            interfaces.append(iface)
        return interfaces

    def _validate_interfaces(self):
        ifaces = self.type_definition.get_value(self.INTERFACES,
                                                self.entity_tpl)
        if ifaces:
            for name, value in ifaces.items():
                if name == 'defaults':
                  self._common_validate_field(
                      value,
                      ['implementation', 'inputs', 'outputs'],
                      'interfaces')
                elif name in (LIFECYCLE, LIFECYCLE_SHORTNAME):
                    self._common_validate_field(
                        value, INTERFACE_DEF_RESERVED_WORDS + InterfacesDef.
                        interfaces_node_lifecycle_operations,
                        'interfaces')
                elif name in (CONFIGURE, CONFIGURE_SHORTNAME):
                    self._common_validate_field(
                        value, INTERFACE_DEF_RESERVED_WORDS + InterfacesDef.
                        interfaces_relationship_configure_operations,
                        'interfaces')
                elif (name in self.type_definition.interfaces
                      or name in self.type_definition.TOSCA_DEF):
                      self._common_validate_field(
                          value,
                          INTERFACE_DEF_RESERVED_WORDS + self._collect_custom_iface_operations(name),
                          'interfaces')
                else:
                    ExceptionCollector.appendException(
                        UnknownFieldError(
                            what='"interfaces" of template "%s"' %
                            self.name, field=name))

    def _collect_custom_iface_operations(self, name):
        allowed_operations = []
        nodetype_iface_def = self.type_definition.interfaces.get(
                              name, self.type_definition.TOSCA_DEF.get(name))
        allowed_operations.extend(nodetype_iface_def.keys())
        if 'type' in nodetype_iface_def:
            iface_type = nodetype_iface_def['type']
            if iface_type in self.type_definition.custom_def:
                iface_type_def = self.type_definition.custom_def[iface_type]
            else:
                iface_type_def = self.type_definition.TOSCA_DEF[iface_type]
            allowed_operations.extend(iface_type_def.keys())
        allowed_operations = [op for op in allowed_operations if
                              op not in INTERFACE_DEF_RESERVED_WORDS]
        return allowed_operations

    def get_capability(self, name):
        """Provide named capability

        :param name: name of capability
        :return: capability object if found, None otherwise
        """
        caps = self.get_capabilities()
        if caps and name in caps.keys():
            return caps[name]



#!/usr/bin/env python
"""
Applies a Unfurl ensemble

For each configuration, run it if required, then record the result
"""
from __future__ import print_function

from .yamlmanifest import runJob
from .support import Status
from . import __version__, initLogging, getHomeConfigPath, DefaultNames
from .init import createProject, cloneEnsemble, createHome
from .util import filterEnv, getPackageDigest
from .localenv import LocalEnv
import click
import sys
import os
import os.path
import traceback
import logging
import functools
import subprocess
import shlex
import json

_latestJobs = []  # for testing


def option_group(*options):
    return lambda func: functools.reduce(lambda a, b: b(a), options, func)


@click.group()
@click.pass_context
@click.option(
    "--home",
    envvar="UNFURL_HOME",
    type=click.Path(exists=False),
    help="Path to .unfurl_home",
)
@click.option("--runtime", envvar="UNFURL_RUNTIME", help="use this runtime")
@click.option(
    "--no-runtime",
    envvar="UNFURL_NORUNTIME",
    default=False,
    is_flag=True,
    help="ignore runtime settings",
)
@click.option("-v", "--verbose", count=True, help="verbose mode (-vvv for more)")
@click.option(
    "-q",
    "--quiet",
    default=False,
    is_flag=True,
    help="Only output errors to the stdout",
)
@click.option(
    "--logfile", default=None, help="Log file for messages during quiet operation"
)
@click.option(
    "--tmp",
    envvar="UNFURL_TMPDIR",
    type=click.Path(exists=True),
    help="Directory for saving temporary files",
)
@click.option("--loglevel", envvar="UNFURL_LOGGING", help="log level (overrides -v)")
def cli(ctx, verbose=0, quiet=False, logfile=None, loglevel=None, tmp=None, **kw):
    # ensure that ctx.obj exists and is a dict (in case `cli()` is called
    # by means other than the `if` block below
    ctx.ensure_object(dict)
    ctx.obj.update(kw)

    if tmp is not None:
        os.environ["UNFURL_TMPDIR"] = tmp

    levels = [logging.INFO, logging.DEBUG, 5, 5, 5]
    if quiet:
        logLevel = logging.CRITICAL
    else:
        # TRACE (5)
        logLevel = levels[min(verbose, 3)]

    if loglevel:  # UNFURL_LOGGING overrides command line
        logLevel = dict(CRITICAL=50, ERROR=40, WARNING=30, INFO=20, DEBUG=10)[
            loglevel.upper()
        ]
    # verbose: 0 == INFO, -1 == CRITICAL, >= 1 == DEBUG
    if logLevel == logging.CRITICAL:
        verbose = -1
    elif logLevel >= logging.INFO:
        verbose = 0
    else: # must be DEBUG
        verbose = 1
    ctx.obj["verbose"] = verbose
    initLogging(logLevel, logfile)


jobControlOptions = option_group(
    click.option(
        "--dryrun",
        default=False,
        is_flag=True,
        help="Do not modify anything, just do a dry run.",
    ),
    click.option(
        "--commit",
        default=False,
        is_flag=True,
        help="Commit modified files to the instance repository. (Default: False)",
    ),
    click.option(
        "--dirty",
        default=False,
        is_flag=True,
        help="Run even if there are uncommitted changes in the instance repository. (Default: False)",
    ),
    click.option("-m", "--message", help="commit message to use"),
    click.option(
        "--jobexitcode",
        type=click.Choice(["error", "degraded", "never"]),
        default="never",
        help="Set exit code to 1 if job status is not ok.",
    ),
)

commonJobFilterOptions = option_group(
    click.option("--template", help="TOSCA template to target"),
    click.option("--instance", help="instance name to target"),
    click.option("--query", help="Run the given expression upon job completion"),
    click.option("--trace", default=0, help="Set the query's trace level"),
    click.option(
        "--output",
        type=click.Choice(["text", "json", "none"]),
        default="text",
        help="How to print summary of job run",
    ),
)


@cli.command(short_help="Record and run an ad-hoc command")
@click.pass_context
# @click.argument("action", default="*:upgrade")
@click.option("--ensemble", default="", type=click.Path(exists=False))
# XXX:
# @click.option(
#     "--append", default=False, is_flag=True, help="add this command to the previous"
# )
# @click.option(
#     "--replace", default=False, is_flag=True, help="replace the previous command"
# )
@jobControlOptions
@commonJobFilterOptions
@click.option("--host", help="host to run the command on")
@click.option("--operation", help="TOSCA operation to run")
@click.option("--module", help="ansible module to run (default: command)")
@click.argument("cmdline", nargs=-1, type=click.UNPROCESSED)
def run(ctx, instance="root", cmdline=None, **options):
    """
    Run an ad-hoc command in the context of the given ensemble.
    Use "--" to separate the given command line, for example:

    > unfurl run -- echo 'hello!'

    If --host or --module is set, the ansible configurator will be used. e.g.:

    > unfurl run --host=example.com -- echo 'hello!'
    """
    options.update(ctx.obj)
    options["instance"] = instance
    options["cmdline"] = cmdline
    return _run(options.pop("ensemble"), options, ctx.info_name)


def _run(ensemble, options, workflow=None):
    if workflow:
        options["workflow"] = workflow

    localEnv = LocalEnv(ensemble, options.get("home"))
    if not options.get("no_runtime"):
        runtime = options.get("runtime")
        if not runtime:
            runtime = localEnv.getEngine()
        if runtime and runtime != ".":
            return _runRemote(runtime, ensemble, options, localEnv)
    return _runLocal(ensemble, options)


def _venv(runtime, env):
    if env is None:
        env = os.environ.copy()
    # see virtualenv activate
    env.pop("PYTHONHOME", None)  # unset if set
    env["VIRTUAL_ENV"] = runtime
    env["PATH"] = os.path.join(runtime, "bin") + os.pathsep + env.get("PATH", "")
    return env


def _remoteCmd(runtime, cmdLine, localEnv):
    context = localEnv.getContext()
    kind, sep, rest = runtime.partition(":")
    if context.get("environment"):
        addOnly = kind == "docker"
        env = filterEnv(localEnv.mapValue(context["environment"]), addOnly=addOnly)
    else:
        env = None

    if kind == "venv":
        return (
            _venv(runtime, env),
            ["python", "-m", "unfurl", "--no-runtime"] + cmdLine,
            False,
        )
    # elif docker: docker $container -it run $cmdline
    else:
        # treat as shell command
        cmd = shlex.split(runtime)
        return env, cmd + ["--no-runtime"] + cmdLine, True


def _runRemote(runtime, ensemble, options, localEnv):
    import logging

    logger = logging.getLogger("unfurl")
    logger.debug('running command remotely on "%s"', runtime)
    cmdLine = sys.argv[1:]
    env, remote, shell = _remoteCmd(runtime, cmdLine, localEnv)
    rv = subprocess.call(remote, env=env, shell=shell)
    if options.get("standalone_mode") is False:
        return rv
    else:
        sys.exit(rv)


def _runLocal(ensemble, options):
    job = runJob(ensemble, options)
    _latestJobs.append(job)
    if not job:
        click.echo("Unable to create job")
    elif job.unexpectedAbort:
        click.echo("Job unexpected aborted")
        if options.get("verbose", 0) > 0:
            raise job.unexpectedAbort
    else:
        jsonSummary = {}
        summary = options.get("output")
        if summary == "text":
            click.echo(job.summary())
        elif summary == "json":
            jsonSummary = job.jsonSummary()

        query = options.get("query")
        if query:
            result = job.runQuery(query, options.get("trace"))
            if summary == "json":
                jsonSummary["query"] = query
                jsonSummary["result"] = result
            else:
                click.echo("query: " + query)
                click.echo(result)
        if jsonSummary:
            click.echo(json.dumps(jsonSummary))

    if not job or (
        "jobexitcode" in options
        and options["jobexitcode"] != "never"
        and Status[options["jobexitcode"]] <= job.status
    ):
        if options.get("standalone_mode") is False:
            return 1
        else:
            sys.exit(1)
    else:
        return 0


# XXX update help text sans "configurations"
deployFilterOptions = option_group(
    click.option(
        "--add", default=True, is_flag=True, help="run newly added configurations"
    ),
    click.option(
        "--update",
        default=True,
        is_flag=True,
        help="run configurations that whose spec has changed but don't require a major version change",
    ),
    click.option(
        "--repair",
        type=click.Choice(["error", "degraded", "missing", "none"]),
        default="error",
        help="re-run configurations that are in an error or degraded state",
    ),
    click.option(
        "--upgrade",
        default=False,
        is_flag=True,
        help="run configurations with major version changes or whose spec has changed",
    ),
    click.option(
        "--force",
        default=False,
        is_flag=True,
        help="(re)run operation regardless of instance's status or state",
    ),
    click.option(
        "--prune",
        default=False,
        is_flag=True,
        help="delete instances that are no longer used",
    ),
)


@cli.command()
@click.pass_context
@click.argument("ensemble", default="", type=click.Path(exists=False))
@commonJobFilterOptions
@deployFilterOptions
@jobControlOptions
def deploy(ctx, ensemble=None, **options):
    """
    Deploy the given ensemble
    """
    options.update(ctx.obj)
    return _run(ensemble, options, ctx.info_name)


@cli.command(short_help="Check the status of each instance")
@click.pass_context
@click.argument("ensemble", default="", type=click.Path(exists=False))
@commonJobFilterOptions
@jobControlOptions
def check(ctx, ensemble=None, **options):
    """
    Check and update the status of the ensemble's instances
    """
    options.update(ctx.obj)
    return _run(ensemble, options, ctx.info_name)


@cli.command(short_help="run the discover workflow")
@click.pass_context
@click.argument("ensemble", default="", type=click.Path(exists=False))
@commonJobFilterOptions
@jobControlOptions
def discover(ctx, ensemble=None, **options):
    """
    Update configuration by probing live instances associated with the ensemble
    """
    options.update(ctx.obj)
    return _run(ensemble, options, ctx.info_name)


@cli.command()
@click.pass_context
@click.argument("ensemble", default="", type=click.Path(exists=False))
@commonJobFilterOptions
@jobControlOptions
def undeploy(ctx, ensemble=None, **options):
    """
    Destroy what was deployed.
    """
    options.update(ctx.obj)
    return _run(ensemble, options, ctx.info_name)


@cli.command()
@click.pass_context
@click.argument("ensemble", default="", type=click.Path(exists=False))
@commonJobFilterOptions
@jobControlOptions
def stop(ctx, ensemble=None, **options):
    """
    Stop running instances.
    """
    options.update(ctx.obj)
    return _run(ensemble, options, ctx.info_name)


@cli.command(short_help="Print the given deployment plan")
@click.pass_context
@click.argument("ensemble", default="", type=click.Path(exists=False))
@commonJobFilterOptions
@deployFilterOptions
@click.option("--workflow", default="deploy", help="plan workflow (default: deploy)")
def plan(ctx, ensemble=None, **options):
    "Print the given deployment plan"
    options.update(ctx.obj)
    options["planOnly"] = True
    # XXX show status and task to run including preview of generated templates, cmds to run etc.
    return _run(ensemble, options)


@cli.command(short_help="Create a new unfurl project or ensemble")
@click.pass_context
@click.argument("projectdir", default=".", type=click.Path(exists=False))
@click.option(
    "--mono", default=False, is_flag=True, help="Create one repository for the project."
)
@click.option(
    "--existing",
    default=False,
    is_flag=True,
    help="Add project to nearest existing repository.",
)
def init(ctx, projectdir, **options):
    """
unfurl init [project] # creates a unfurl project with new spec and instance repos
"""
    options.update(ctx.obj)

    if os.path.exists(projectdir):
        if not os.path.isdir(projectdir):
            raise click.ClickException(
                'Can not create project in "' + projectdir + '": file already exists'
            )
        else:
            # destination is inside an existing project, just create a new ensemble
            if os.path.exists(os.path.join(projectdir, DefaultNames.LocalConfig)):
                message = cloneEnsemble(projectdir, projectdir, **options)
                click.echo(message)
                return
            elif os.listdir(projectdir):
                raise click.ClickException(
                    'Can not create project in "'
                    + projectdir
                    + '": folder is not empty'
                )

    homePath, projectPath = createProject(projectdir, **options)
    if homePath:
        click.echo("unfurl home created at %s" % homePath)
    click.echo("New Unfurl project created at %s" % projectPath)


# XXX add --upgrade option
@cli.command(short_help="Manage the unfurl home project")
@click.pass_context
@click.option(
    "--render",
    default=False,
    is_flag=True,
    help="Generate files only (don't create repository)",
)
@click.option("--init", default=False, is_flag=True, help="Create a new home project")
@click.option(
    "--replace",
    default=False,
    is_flag=True,
    help="Replace (and backup) current home project",
)
def home(ctx, init=False, render=False, replace=False, **options):
    options.update(ctx.obj)
    if not render and not init:
        # just display the current home location
        click.echo(getHomeConfigPath(options.get("home")))
        return

    homePath = createHome(render=render, replace=replace, **options)
    action = "rendered" if render else "created"
    if homePath:
        click.echo("unfurl home %s at %s" % (action, homePath))
    else:
        currentHome = getHomeConfigPath(options.get("home"))
        if currentHome:
            click.echo("Can't %s home, it already exists at %s" % (action, currentHome))
        else:
            click.echo("Error: home path is empty")


@cli.command(short_help="Clone a project, ensemble or service template")
@click.pass_context
@click.argument(
    "source",
    # help="path to a service template or ensemble",
)
@click.argument(
    "dest",
    type=click.Path(exists=False),
    default=".",  # , help="path to the new ensemble"
)
@click.option(
    "--mono", default=False, is_flag=True, help="Create one repository for the project."
)
@click.option(
    "--existing",
    default=False,
    is_flag=True,
    help="Add project to nearest existing repository.",
)
def clone(ctx, source, dest, **options):
    """Create a new ensemble or project from a service template or an existing ensemble or project."""
    options.update(ctx.obj)

    message = cloneEnsemble(source, dest, **options)
    click.echo(message)


@cli.command(
    short_help="Run a git command across all repositories",
    context_settings={"ignore_unknown_options": True},
)
@click.pass_context
@click.option(
    "--dir", default=".", type=click.Path(exists=True), help="path to spec repository"
)
@click.argument("gitargs", nargs=-1)
def git(ctx, gitargs, dir="."):
    """
unfurl git --dir=/path/to/start [gitoptions] [gitcmd] [gitcmdoptions]: Runs command on each project repository.
"""
    localEnv = LocalEnv(dir, ctx.obj.get("home"))
    repos = localEnv.getRepos()
    status = 0
    if not repos:
        click.echo("Can't run git command, no repositories found")
    for repo in repos:
        repoPath = os.path.relpath(repo.workingDir, os.path.abspath(dir))
        click.echo("*** Running 'git %s' in './%s'" % (" ".join(gitargs), repoPath))
        _status, stdout, stderr = repo.runCmd(gitargs)
        click.echo(stdout + "\n")
        if stderr.strip():
            click.echo(stderr + "\n", color="red")
        if _status != 0:
            status = _status

    return status


@cli.command()
@click.option(
    "--semver",
    default=False,
    is_flag=True,
    help="Print only the semantic version",
)
def version(semver=False):
    "Print the current version"
    if semver:
        click.echo(__version__())
    else:
        click.echo("unfurl version %s (%s)" % (__version__(True), getPackageDigest()))


@cli.command()
@click.pass_context
@click.argument("cmd", nargs=1, default="")
def help(ctx, cmd=""):
    "Get help on a command"
    if not cmd:
        click.echo(cli.get_help(ctx.parent), color=ctx.color)
        return

    command = ctx.parent.command.commands.get(cmd)
    if command:
        ctx.info_name = cmd  # hack
        click.echo(command.get_help(ctx), color=ctx.color)
    else:
        raise click.UsageError("no help found for unknown command '%s'" % cmd, ctx=ctx)


def main():
    obj = {"standalone_mode": False}
    try:
        rv = cli(standalone_mode=False, obj=obj)
        sys.exit(rv or 0)
    except click.Abort:
        click.secho("Aborted!", fg="red", err=True)
        sys.exit(1)
    except click.ClickException as e:
        if obj.get("verbose", 0) > 0:
            traceback.print_exc(file=sys.stderr)
        click.secho("Error: %s" % e.format_message(), fg="red", err=True)
        sys.exit(e.exit_code)
    except Exception as err:
        if obj.get("verbose", 0) > 0:
            traceback.print_exc(file=sys.stderr)
        else:
            click.secho("Exiting with error: " + str(err), fg="red", err=True)
        sys.exit(1)


def vaultclient():
    try:
        localEnv = LocalEnv(".")
    except Exception as err:
        click.echo(str(err), err=True)
        return 1

    print(localEnv.getVaultPassword() or "")
    return 0


if __name__ == "__main__":
    main()



import os.path
from unfurl.eval import Ref, mapValue
from unfurl.support import abspath, getdir
from unfurl.util import which
from jinja2.filters import contextfilter

# from ansible.errors import AnsibleError, AnsibleFilterError


@contextfilter
def ref(context, ref, **vars):
    refContext = context["__unfurl"]
    return Ref(ref, vars=vars).resolveOne(refContext)


@contextfilter
def mapValueFilter(context, ref, **vars):
    refContext = context["__unfurl"]
    if vars:
        refContext = refContext.copy(vars=vars)
    return mapValue(ref, refContext)


@contextfilter
def _abspath(context, path, relativeTo=None, mkdir=True):
    """
    {{ 'foo' | abspath }}

    or

    {{ 'foo' | abspath('local') }}
    """
    refContext = context["__unfurl"]
    return abspath(refContext, path, relativeTo, mkdir)


@contextfilter
def get_dir(context, relativeTo, mkdir=True):
    refContext = context["__unfurl"]
    return getdir(refContext, relativeTo, mkdir)


class FilterModule(object):
    def filters(self):
        return {
            "ref": ref,
            "eval": ref,
            "mapValue": mapValueFilter,
            "abspath": _abspath,
            "get_dir": get_dir,
            "which": which,
        }
